# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test of Policy Engine For Tacker"""

import fixtures
import io
from unittest import mock
from urllib import request as urlrequest

from oslo_policy import policy as common_policy
from oslo_serialization import jsonutils as json
from oslo_utils import importutils

import tacker
from tacker.api.v1 import attributes
from tacker.common import exceptions
from tacker import context
from tacker import manager
from tacker import policy
from tacker.tests import base


class PolicyFileTestCase(base.BaseTestCase):
    def setUp(self):
        super(PolicyFileTestCase, self).setUp()
        self.skipTest("Not ready yet")
        policy.reset()
        self.addCleanup(policy.reset)
        self.context = context.Context('fake', 'fake', is_admin=False)
        self.target = {}
        self.tempdir = self.useFixture(fixtures.TempDir())

    def test_modified_policy_reloads(self):
        def fake_find_config_file(_1, _2):
            return self.tempdir.join('policy')

        with mock.patch.object(tacker.common.utils,
                               'find_config_file',
                               new=fake_find_config_file):
            tmpfilename = fake_find_config_file(None, None)
            action = "example:test"
            with open(tmpfilename, "w") as policyfile:
                policyfile.write("""{"example:test": ""}""")
            policy.init()
            policy.enforce(self.context, action, self.target)
            with open(tmpfilename, "w") as policyfile:
                policyfile.write("""{"example:test": "!"}""")
            # NOTE(vish): reset stored policy cache so we don't have to
            # sleep(1)
            policy._POLICY_CACHE = {}
            policy.init()
            self.assertRaises(exceptions.PolicyNotAuthorized,
                              policy.enforce,
                              self.context,
                              action,
                              self.target)


class PolicyTestCase(base.BaseTestCase):
    def setUp(self):
        super(PolicyTestCase, self).setUp()
        self.skipTest("Not ready yet")
        policy.reset()
        self.addCleanup(policy.reset)
        # NOTE(vish): preload rules to circumvent reloading from file
        policy.init()
        rules = {
            "true": '@',
            "example:allowed": '@',
            "example:denied": '!',
            "example:get_http": "http:http://www.example.com",
            "example:my_file": "role:compute_admin or tenant_id:%(tenant_id)s",
            "example:early_and_fail": "! and @",
            "example:early_or_success": "@ or !",
            "example:lowercase_admin": "role:admin or role:sysadmin",
            "example:uppercase_admin": "role:ADMIN or role:sysadmin",
        }
        # NOTE(vish): then overload underlying rules
        common_policy.set_rules(common_policy.Rules(
            dict((k, common_policy.parse_rule(v))
                 for k, v in rules.items())))
        self.context = context.Context('fake', 'fake', roles=['member'])
        self.target = {}

    def test_enforce_nonexistent_action_throws(self):
        action = "example:noexist"
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, self.target)

    def test_enforce_bad_action_throws(self):
        action = "example:denied"
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, self.target)

    def test_check_bad_action_noraise(self):
        action = "example:denied"
        result = policy.check(self.context, action, self.target)
        self.assertFalse(result)

    def test_check_non_existent_action(self):
        action = "example:idonotexist"
        result_1 = policy.check(self.context, action, self.target)
        self.assertFalse(result_1)
        result_2 = policy.check(self.context, action, self.target,
                                might_not_exist=True)
        self.assertTrue(result_2)

    def test_enforce_good_action(self):
        action = "example:allowed"
        result = policy.enforce(self.context, action, self.target)
        self.assertEqual(True, result)

    def test_enforce_http_true(self):

        def fakeurlopen(url, post_data):
            return io.StringIO("True")

        with mock.patch.object(urlrequest, 'urlopen', new=fakeurlopen):
            action = "example:get_http"
            target = {}
            result = policy.enforce(self.context, action, target)
            self.assertEqual(True, result)

    def test_enforce_http_false(self):

        def fakeurlopen(url, post_data):
            return io.StringIO("False")

        with mock.patch.object(urlrequest, 'urlopen', new=fakeurlopen):
            action = "example:get_http"
            target = {}
            self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                              self.context, action, target)

    def test_templatized_enforcement(self):
        target_mine = {'tenant_id': 'fake'}
        target_not_mine = {'tenant_id': 'another'}
        action = "example:my_file"
        policy.enforce(self.context, action, target_mine)
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, target_not_mine)

    def test_early_AND_enforcement(self):
        action = "example:early_and_fail"
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, self.target)

    def test_early_OR_enforcement(self):
        action = "example:early_or_success"
        policy.enforce(self.context, action, self.target)

    def test_ignore_case_role_check(self):
        lowercase_action = "example:lowercase_admin"
        uppercase_action = "example:uppercase_admin"
        # NOTE(dprince) we mix case in the Admin role here to ensure
        # case is ignored
        admin_context = context.Context('admin', 'fake', roles=['AdMiN'])
        policy.enforce(admin_context, lowercase_action, self.target)
        policy.enforce(admin_context, uppercase_action, self.target)


class DefaultPolicyTestCase(base.BaseTestCase):

    def setUp(self):
        super(DefaultPolicyTestCase, self).setUp()
        self.skipTest("Not ready yet")
        policy.reset()
        policy.init()
        self.addCleanup(policy.reset)

        self.rules = {
            "default": '',
            "example:exist": '!',
        }

        self._set_rules('default')

        self.context = context.Context('fake', 'fake')

    def _set_rules(self, default_rule):
        rules = common_policy.Rules(
            dict((k, common_policy.parse_rule(v))
                 for k, v in self.rules.items()), default_rule)
        common_policy.set_rules(rules)

    def test_policy_called(self):
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, "example:exist", {})

    def test_not_found_policy_calls_default(self):
        policy.enforce(self.context, "example:noexist", {})

    def test_default_not_found(self):
        self._set_rules("default_noexist")
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, "example:noexist", {})


FAKE_RESOURCE_NAME = 'something'
FAKE_RESOURCE = {"%ss" % FAKE_RESOURCE_NAME:
                 {'attr': {'allow_post': True,
                           'allow_put': True,
                           'is_visible': True,
                           'default': None,
                           'enforce_policy': True,
                           'validate': {'type:dict':
                                        {'sub_attr_1': {'type:string': None},
                                         'sub_attr_2': {'type:string': None}}}
                           }}}


class TackerPolicyTestCase(base.BaseTestCase):

    def setUp(self):
        super(TackerPolicyTestCase, self).setUp()
        self.skipTest("Not ready yet")
        policy.reset()
        policy.init()
        self.addCleanup(policy.reset)
        self.admin_only_legacy = "role:admin"
        self.admin_or_owner_legacy = "role:admin or tenant_id:%(tenant_id)s"
        # Add a Fake 'something' resource to RESOURCE_ATTRIBUTE_MAP
        attributes.RESOURCE_ATTRIBUTE_MAP.update(FAKE_RESOURCE)
        self.rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            "context_is_admin": "role:admin",
            "admin_or_network_owner": "rule:context_is_admin or "
                                      "tenant_id:%(network:tenant_id)s",
            "admin_or_owner": ("rule:context_is_admin or "
                               "tenant_id:%(tenant_id)s"),
            "admin_only": "rule:context_is_admin",
            "regular_user": "role:user",
            "shared": "field:networks:shared=True",
            "external": "field:networks:router:external=True",
            "default": '@',

            "create_network": "rule:admin_or_owner",
            "create_network:shared": "rule:admin_only",
            "update_network": '@',
            "update_network:shared": "rule:admin_only",

            "get_network": "rule:admin_or_owner or "
                           "rule:shared or "
                           "rule:external",
            "create_port:mac": "rule:admin_or_network_owner",
            "create_something": "rule:admin_or_owner",
            "create_something:attr": "rule:admin_or_owner",
            "create_something:attr:sub_attr_1": "rule:admin_or_owner",
            "create_something:attr:sub_attr_2": "rule:admin_only",

            "get_firewall_policy": "rule:admin_or_owner or "
                            "rule:shared",
            "get_firewall_rule": "rule:admin_or_owner or "
                            "rule:shared"
        }.items())

        def fakepolicyinit():
            common_policy.set_rules(common_policy.Rules(self.rules))

        def remove_fake_resource():
            del attributes.RESOURCE_ATTRIBUTE_MAP["%ss" % FAKE_RESOURCE_NAME]

        self.patcher = mock.patch.object(tacker.policy,
                                         'init',
                                         new=fakepolicyinit)
        self.patcher.start()
        self.addCleanup(remove_fake_resource)
        self.context = context.Context('fake', 'fake', roles=['user'])
        plugin_klass = importutils.import_class(
            "tacker.db.db_base_plugin_v2.TackerDbPluginV2")
        self.manager_patcher = mock.patch('tacker.manager.TackerManager')
        fake_manager = self.manager_patcher.start()
        fake_manager_instance = fake_manager.return_value
        fake_manager_instance.plugin = plugin_klass()

    def _test_action_on_attr(self, context, action, attr, value,
                             exception=None):
        action = "%s_network" % action
        target = {'tenant_id': 'the_owner', attr: value}
        if exception:
            self.assertRaises(exception, policy.enforce,
                              context, action, target)
        else:
            result = policy.enforce(context, action, target)
            self.assertEqual(True, result)

    def _test_nonadmin_action_on_attr(self, action, attr, value,
                                      exception=None):
        user_context = context.Context('', "user", roles=['user'])
        self._test_action_on_attr(user_context, action, attr,
                                  value, exception)

    def test_nonadmin_write_on_private_fails(self):
        self._test_nonadmin_action_on_attr('create', 'shared', False,
                                           exceptions.PolicyNotAuthorized)

    def test_nonadmin_read_on_private_fails(self):
        self._test_nonadmin_action_on_attr('get', 'shared', False,
                                           exceptions.PolicyNotAuthorized)

    def test_nonadmin_write_on_shared_fails(self):
        self._test_nonadmin_action_on_attr('create', 'shared', True,
                                           exceptions.PolicyNotAuthorized)

    def test_nonadmin_read_on_shared_succeeds(self):
        self._test_nonadmin_action_on_attr('get', 'shared', True)

    def _test_enforce_adminonly_attribute(self, action):
        admin_context = context.get_admin_context()
        target = {'shared': True}
        result = policy.enforce(admin_context, action, target)
        self.assertEqual(True, result)

    def test_enforce_adminonly_attribute_create(self):
        self._test_enforce_adminonly_attribute('create_network')

    def test_enforce_adminonly_attribute_update(self):
        self._test_enforce_adminonly_attribute('update_network')

    def test_enforce_adminonly_attribute_no_context_is_admin_policy(self):
        del self.rules[policy.ADMIN_CTX_POLICY]
        self.rules['admin_only'] = common_policy.parse_rule(
            self.admin_only_legacy)
        self.rules['admin_or_owner'] = common_policy.parse_rule(
            self.admin_or_owner_legacy)
        self._test_enforce_adminonly_attribute('create_network')

    def test_enforce_adminonly_attribute_nonadminctx_returns_403(self):
        action = "create_network"
        target = {'shared': True, 'tenant_id': 'somebody_else'}
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, target)

    def test_enforce_adminonly_nonadminctx_no_ctx_is_admin_policy_403(self):
        del self.rules[policy.ADMIN_CTX_POLICY]
        self.rules['admin_only'] = common_policy.parse_rule(
            self.admin_only_legacy)
        self.rules['admin_or_owner'] = common_policy.parse_rule(
            self.admin_or_owner_legacy)
        action = "create_network"
        target = {'shared': True, 'tenant_id': 'somebody_else'}
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, target)

    def _test_build_subattribute_match_rule(self, validate_value):
        bk = FAKE_RESOURCE['%ss' % FAKE_RESOURCE_NAME]['attr']['validate']
        FAKE_RESOURCE['%ss' % FAKE_RESOURCE_NAME]['attr']['validate'] = (
            validate_value)
        action = "create_something"
        target = {'tenant_id': 'fake', 'attr': {'sub_attr_1': 'x'}}
        self.assertFalse(policy._build_subattr_match_rule(
            'attr',
            FAKE_RESOURCE['%ss' % FAKE_RESOURCE_NAME]['attr'],
            action,
            target))
        FAKE_RESOURCE['%ss' % FAKE_RESOURCE_NAME]['attr']['validate'] = bk

    def test_build_subattribute_match_rule_empty_dict_validator(self):
        self._test_build_subattribute_match_rule({})

    def test_build_subattribute_match_rule_wrong_validation_info(self):
        self._test_build_subattribute_match_rule(
            {'type:dict': 'wrong_stuff'})

    def test_enforce_subattribute(self):
        action = "create_something"
        target = {'tenant_id': 'fake', 'attr': {'sub_attr_1': 'x'}}
        result = policy.enforce(self.context, action, target, None)
        self.assertEqual(True, result)

    def test_enforce_admin_only_subattribute(self):
        action = "create_something"
        target = {'tenant_id': 'fake', 'attr': {'sub_attr_1': 'x',
                                                'sub_attr_2': 'y'}}
        result = policy.enforce(context.get_admin_context(),
                                action, target, None)
        self.assertEqual(True, result)

    def test_enforce_admin_only_subattribute_nonadminctx_returns_403(self):
        action = "create_something"
        target = {'tenant_id': 'fake', 'attr': {'sub_attr_1': 'x',
                                                'sub_attr_2': 'y'}}
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, target, None)

    def test_enforce_regularuser_on_read(self):
        action = "get_network"
        target = {'shared': True, 'tenant_id': 'somebody_else'}
        result = policy.enforce(self.context, action, target)
        self.assertTrue(result)

    def test_enforce_firewall_policy_shared(self):
        action = "get_firewall_policy"
        target = {'shared': True, 'tenant_id': 'somebody_else'}
        result = policy.enforce(self.context, action, target)
        self.assertTrue(result)

    def test_enforce_firewall_rule_shared(self):
        action = "get_firewall_rule"
        target = {'shared': True, 'tenant_id': 'somebody_else'}
        result = policy.enforce(self.context, action, target)
        self.assertTrue(result)

    def test_enforce_tenant_id_check(self):
        # Trigger a policy with rule admin_or_owner
        action = "create_network"
        target = {'tenant_id': 'fake'}
        result = policy.enforce(self.context, action, target)
        self.assertTrue(result)

    def test_enforce_tenant_id_check_parent_resource(self):

        def fakegetnetwork(*args, **kwargs):
            return {'tenant_id': 'fake'}

        action = "create_port:mac"
        with mock.patch.object(manager.TackerManager.get_instance().plugin,
                               'get_network', new=fakegetnetwork):
            target = {'network_id': 'whatever'}
            result = policy.enforce(self.context, action, target)
            self.assertTrue(result)

    def test_enforce_plugin_failure(self):

        def fakegetnetwork(*args, **kwargs):
            raise NotImplementedError('Blast!')

        # the policy check and plugin method we use in this test are irrelevant
        # so long that we verify that, if *f* blows up, the behavior of the
        # policy engine to propagate the exception is preserved
        action = "create_port:mac"
        with mock.patch.object(manager.TackerManager.get_instance().plugin,
                               'get_network', new=fakegetnetwork):
            target = {'network_id': 'whatever'}
            self.assertRaises(NotImplementedError,
                              policy.enforce,
                              self.context,
                              action,
                              target)

    def test_enforce_tenant_id_check_parent_resource_bw_compatibility(self):

        def fakegetnetwork(*args, **kwargs):
            return {'tenant_id': 'fake'}

        del self.rules['admin_or_network_owner']
        self.rules['admin_or_network_owner'] = common_policy.parse_rule(
            "role:admin or tenant_id:%(network_tenant_id)s")
        action = "create_port:mac"
        with mock.patch.object(manager.TackerManager.get_instance().plugin,
                               'get_network', new=fakegetnetwork):
            target = {'network_id': 'whatever'}
            result = policy.enforce(self.context, action, target)
            self.assertTrue(result)

    def test_tenant_id_check_no_target_field_raises(self):
        # Try and add a bad rule
        self.assertRaises(
            exceptions.PolicyInitError,
            common_policy.parse_rule,
            'tenant_id:(wrong_stuff)')

    def _test_enforce_tenant_id_raises(self, bad_rule):
        self.rules['admin_or_owner'] = common_policy.parse_rule(bad_rule)
        # Trigger a policy with rule admin_or_owner
        action = "create_network"
        target = {'tenant_id': 'fake'}
        policy.init()
        self.assertRaises(exceptions.PolicyCheckError,
                          policy.enforce,
                          self.context, action, target)

    def test_enforce_tenant_id_check_malformed_target_field_raises(self):
        self._test_enforce_tenant_id_raises('tenant_id:%(malformed_field)s')

    def test_enforce_tenant_id_check_invalid_parent_resource_raises(self):
        self._test_enforce_tenant_id_raises('tenant_id:%(foobaz_tenant_id)s')

    def test_get_roles_context_is_admin_rule_missing(self):
        rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            "some_other_rule": "role:admin",
        }.items())
        common_policy.set_rules(common_policy.Rules(rules))
        # 'admin' role is expected for bw compatibility
        self.assertEqual(['admin'], policy.get_admin_roles())

    def test_get_roles_with_role_check(self):
        rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            policy.ADMIN_CTX_POLICY: "role:admin",
        }.items())
        common_policy.set_rules(common_policy.Rules(rules))
        self.assertEqual(['admin'], policy.get_admin_roles())

    def test_get_roles_with_rule_check(self):
        rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            policy.ADMIN_CTX_POLICY: "rule:some_other_rule",
            "some_other_rule": "role:admin",
        }.items())
        common_policy.set_rules(common_policy.Rules(rules))
        self.assertEqual(['admin'], policy.get_admin_roles())

    def test_get_roles_with_or_check(self):
        self.rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            policy.ADMIN_CTX_POLICY: "rule:rule1 or rule:rule2",
            "rule1": "role:admin_1",
            "rule2": "role:admin_2"
        }.items())
        self.assertEqual(['admin_1', 'admin_2'],
                         policy.get_admin_roles())

    def test_get_roles_with_other_rules(self):
        self.rules = dict((k, common_policy.parse_rule(v)) for k, v in {
            policy.ADMIN_CTX_POLICY: "role:xxx or other:value",
        }.items())
        self.assertEqual(['xxx'], policy.get_admin_roles())

    def _test_set_rules_with_deprecated_policy(self, input_rules,
                                               expected_rules):
        policy._set_rules(json.dumps(input_rules))
        # verify deprecated policy has been removed
        for pol in input_rules:
            self.assertNotIn(pol, common_policy._rules)
        # verify deprecated policy was correctly translated. Iterate
        # over items for compatibility with unittest2 in python 2.6
        for rule in expected_rules:
            self.assertIn(rule, common_policy._rules)
            self.assertEqual(expected_rules[rule],
                             str(common_policy._rules[rule]))

    def test_set_rules_with_deprecated_view_policy(self):
        self._test_set_rules_with_deprecated_policy(
            {'extension:router:view': 'rule:admin_or_owner'},
            {'get_network:router:external': 'rule:admin_or_owner'})

    def test_set_rules_with_deprecated_set_policy(self):
        expected_policies = ['create_network:provider:network_type',
                             'create_network:provider:physical_network',
                             'create_network:provider:segmentation_id',
                             'update_network:provider:network_type',
                             'update_network:provider:physical_network',
                             'update_network:provider:segmentation_id']
        self._test_set_rules_with_deprecated_policy(
            {'extension:provider_network:set': 'rule:admin_only'},
            dict((policy, 'rule:admin_only') for policy in
                 expected_policies))



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glance_store
from oslo_config import cfg

from tacker.conf import conductor
from tacker.conf import coordination
from tacker.conf import vnf_lcm
from tacker.conf import vnf_package

CONF = cfg.CONF
CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')

vnf_package.register_opts(CONF)
vnf_lcm.register_opts(CONF)
conductor.register_opts(CONF)
coordination.register_opts(CONF)
glance_store.register_opts(CONF)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import codecs
from datetime import datetime
import os
import yaml


DUMMY_NS_2_NAME = 'dummy_ns_2'


def _get_template(name):
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "../vnfm/infra_drivers/openstack/data/", name)
    f = codecs.open(filename, encoding='utf-8', errors='strict')
    return f.read()


tosca_cvnf_vnfd = _get_template('test_tosca_cvnf.yaml')
tosca_vnfd_openwrt = _get_template('test_tosca_openwrt.yaml')
tosca_vnfd_openwrt_param = _get_template('test_tosca_openwrt_param.yaml')
tosca_invalid_vnfd = _get_template('test_tosca_parser_failure.yaml')
etsi_vnfd = _get_template('etsi_nfv/tosca_vnfd.yaml')
etsi_vnfd_group = _get_template('etsi_nfv/tosca_vnfd_group_member.yaml')
config_data = _get_template('config_data.yaml')
update_config_data = _get_template('update_config_data.yaml')
hot_data = _get_template('hot_data.yaml')
param_data = _get_template('param_data.yaml')
update_param_data = _get_template('update_param_data.yaml')
update_invalid_param_data = _get_template('update_invalid_param_data.yaml')
update_new_param_data = _get_template('update_new_param_data.yaml')
vnffg_params = _get_template('vnffg_params.yaml')
vnffg_multi_params = _get_template('vnffg_multi_params.yaml')
vnffgd_template = yaml.safe_load(_get_template('vnffgd_template.yaml'))
vnffgd_tosca_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_template.yaml'))
vnffgd_tosca_no_classifier_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_no_classifier_template.yaml'))
vnffgd_tosca_template_for_update = yaml.safe_load(_get_template(
    'tosca_vnffgd_template_for_update.yaml'))
vnffgd_legacy_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_legacy_template_for_update.yaml'))
vnffgd_tosca_param_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_param_template.yaml'))
vnffgd_tosca_str_param_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_str_param_template.yaml'))
vnffgd_tosca_multi_param_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_multi_param_template.yaml'))
vnffgd_invalid_tosca_template = yaml.safe_load(_get_template(
    'tosca_invalid_vnffgd_template.yaml'))
vnffgd_tosca_dupl_criteria_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_dupl_criteria_template.yaml'))
vnfd_scale_tosca_template = _get_template('tosca_scale.yaml')
vnfd_alarm_respawn_tosca_template = _get_template(
    'test_tosca_vnfd_alarm_respawn.yaml')
vnfd_alarm_scale_tosca_template = _get_template(
    'test_tosca_vnfd_alarm_scale.yaml')
vnfd_alarm_multi_actions_tosca_template = _get_template(
    'test_tosca_vnfd_alarm_multi_actions.yaml')
nsd_tosca_template = yaml.safe_load(_get_template('tosca_nsd_template.yaml'))
vnffgd_wrong_cp_number_template = yaml.safe_load(_get_template(
    'tosca_vnffgd_wrong_cp_number_template.yaml'))
vnfd_instance_reservation_alarm_scale_tosca_template = _get_template(
    'test_tosca-vnfd-instance-reservation.yaml')
hot_grant = _get_template('hot_grant.yaml')
hot_scale_grant = _get_template('hot_scale_grant.yaml')
hot_scale_nest_grant = _get_template('hot_scale_nest_grant.yaml')
hot_scale_initial = _get_template('hot_scale_initial.yaml')
hot_scale_nest_initial = _get_template('hot_scale_nest_initial.yaml')


def get_dummy_vnfd_obj():
    return {'vnfd': {'service_types': [{'service_type': 'vnfd'}],
                     'name': 'dummy_vnfd',
                     'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                     'attributes': {'vnfd': yaml.safe_load(
                         tosca_vnfd_openwrt)},
                     'description': 'dummy_vnfd_description',
                     'template_source': 'onboarded',
            'auth': {'tenantName': 'admin', 'passwordCredentials': {
                'username': 'admin', 'password': 'devstack'}}}}


def get_invalid_vnfd_obj():
    return {'vnfd': {'service_types': [{'service_type': 'vnfd'}],
                     'name': 'dummy_vnfd',
                     'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                     'attributes': {'vnfd': yaml.safe_load(
                         tosca_invalid_vnfd)},
                     'description': 'dummy_vnfd_description',
                     'template_source': 'onboarded',
            'auth': {'tenantName': 'admin', 'passwordCredentials': {
                'username': 'admin', 'password': 'devstack'}}}}


def get_dummy_vnfd_obj_inline():
    return {'vnfd': {'service_types': [{'service_type': 'vnfd'}],
                     'name': 'tmpl-koeak4tqgoqo8cr4-dummy_inline_vnf',
                     'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                     'attributes': {'vnfd': yaml.safe_load(
                         tosca_vnfd_openwrt)},
                     'template_source': 'inline',
            'auth': {'tenantName': 'admin', 'passwordCredentials': {
                'username': 'admin', 'password': 'devstack'}}}}


def get_dummy_inline_vnf_obj():
    return {'vnf': {'description': 'dummy_inline_vnf_description',
                    'vnfd_template': yaml.safe_load(tosca_vnfd_openwrt),
                    'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                    'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                    'name': 'dummy_inline_vnf',
                    'attributes': {},
                    'vnfd_id': None}}


def get_dummy_inline_cvnf_obj():
    return {'vnf': {'description': 'dummy_inline_cvnf_description',
                    'vnfd_template': yaml.safe_load(tosca_cvnf_vnfd),
                    'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                    'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                    'name': 'dummy_cvnf',
                    'attributes': {},
                    'vnfd_id': None}}


def get_dummy_vnf_obj():
    return {'vnf': {'description': 'dummy_vnf_description',
                    'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                    'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                    'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                    'name': 'dummy_vnf',
                    'deleted_at': datetime.min,
                    'attributes': {},
                    'vnfd_template': None}}


def get_dummy_vnf_config_obj():
    return {'vnf': {'attributes': {'config': {'vdus': {'vdu1': {
        'config': {'firewall': 'dummy_firewall_values'}}}}}}}


def get_dummy_vnf_invalid_config_type_obj():
    return {'vnf': {'attributes': {'config': 'dummy_config'}}}


def get_dummy_vnf_invalid_param_content():
    return {'vnf': {'attributes': {'param_values': {}}}}


def get_dummy_vnf_param_obj():
    return {'vnf': {'attributes': {'param_values':
        {'flavor': 'm1.tiny',
         'reservation_id': '99999999-3925-4c9e-9074-239a902b68d7'}}}}


def get_dummy_vnf_invalid_param_type_obj():
    return {'vnf': {'attributes': {'param_values': 'dummy_param'}}}


def get_dummy_vnf(status='PENDING_CREATE', scaling_group=False,
                  instance_id=None):
    dummy_vnf = {'status': status, 'instance_id': instance_id, 'name':
        'test_openwrt', 'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
        'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
        'vnfd': {
            'service_types': [{'service_type': 'vnfd',
            'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
            'description': 'OpenWRT with services',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'mgmt_driver': 'openwrt',
            'attributes': {'vnfd': tosca_vnfd_openwrt},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec',
            'name': 'openwrt_services'},
        'mgmt_ip_address': None, 'service_context': [],
        'attributes': {'param_values': ''},
        'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
        'description': 'OpenWRT with services'}
    if scaling_group:
        dummy_vnf['attributes'].update({'scaling_group_names':
                                   '{"SP1": "SP1_group"}',
                                   'heat_template': 'test'})
    return dummy_vnf


def get_dummy_vnf_test(status='PENDING_CREATE', scaling_group=False,
                  instance_id=None):
    dummy_vnf = {'status': status, 'instance_id': instance_id, 'name':
        'test_openwrt', 'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
        'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
        'vnfd': {
            'service_types': [{'service_type': 'vnfd',
            'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
            'description': 'OpenWRT with services',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'mgmt_driver': 'openwrt',
            'attributes': {'vnfd_simple': tosca_vnfd_openwrt},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec',
            'name': 'openwrt_services'},
        'mgmt_ip_address': None, 'service_context': [],
        'attributes': {'param_values': ''},
        'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
        'description': 'OpenWRT with services'}
    if scaling_group:
        dummy_vnf['attributes'].update({'scaling_group_names':
                                   '{"SP1": "SP1_group"}',
                                   'heat_template': 'test'})
    return dummy_vnf


def get_dummy_vnf_etsi(status='PENDING_CREATE', scaling_group=False,
                       instance_id=None, flavour='Simple', vnfd_name=None):
    vnfd_key = 'vnfd_' + flavour
    dummy_vnf = {'status': status, 'instance_id': instance_id, 'name':
        'test_openwrt', 'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
        'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
        'vnfd': {
            'service_types': [{'service_type': 'vnfd',
            'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
            'description': 'OpenWRT with services',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'mgmt_driver': 'openwrt',
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec',
            'name': 'openwrt_services'},
        'mgmt_ip_address': None, 'service_context': [],
        'attributes': {'param_values': ''},
        'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
        'description': 'OpenWRT with services'}
    if not vnfd_name:
        # Set vnfd including without "tosca.groups.nfv.PlacementGroup"
        dummy_vnf['vnfd']['attributes'] = {vnfd_key: etsi_vnfd}
    else:
        # Set vnfd including with "tosca.groups.nfv.PlacementGroup"
        dummy_vnf['vnfd']['attributes'] = {vnfd_key: etsi_vnfd_group}
    if scaling_group:
        dummy_vnf['attributes'].update({'scaling_group_names':
                                   '{"SP1": "SP1_group"}',
                                   'heat_template': 'test'})
    return dummy_vnf


def get_dummy_vnf_config_attr():
    return {'status': 'PENDING_CREATE', 'instance_id': None, 'name':
        'test_openwrt', 'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
        'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
        'vnfd': {'service_types': [{'service_type': 'vnfd',
            'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
            'description': 'OpenWRT with services',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'mgmt_driver': 'openwrt',
            'attributes': {'vnfd': tosca_vnfd_openwrt},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec', 'name':
            'openwrt_services'}, 'mgmt_ip_address': None,
            'service_context': [],
            'attributes': {'config': config_data},
            'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
            'description': 'OpenWRT with services'}


def get_dummy_vnf_param_attr():
    return {'status': 'PENDING_CREATE', 'instance_id': None, 'name':
        'test_openwrt', 'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
        'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
        'vnfd': {'service_types': [{'service_type': 'vnfd',
            'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
            'description': 'OpenWRT with services',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'mgmt_driver': 'openwrt',
            'attributes': {'vnfd': tosca_vnfd_openwrt_param},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec',
            'name': 'openwrt_services'},
        'mgmt_url': None, 'service_context': [],
        'attributes': {'heat_template': hot_data,
                       'param_values': param_data},
        'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
        'description': 'OpenWRT with services'}


def get_dummy_vnf_update_config():
    return {'vnf': {'attributes': {'config': update_config_data}}}


def get_dummy_vnf_update_param():
    return {'vnf': {'attributes': {'param_values': update_param_data}}}


def get_dummy_vnf_update_new_param():
    return {'vnf': {'attributes': {'param_values': update_new_param_data}}}


def get_dummy_vnf_update_invalid_param():
    return {'vnf': {'attributes': {'param_values': update_invalid_param_data}}}


def get_dummy_vnf_update_empty_param():
    return {'vnf': {'attributes': {'param_values': {}}}}


def get_vim_obj():
    return {'vim': {'type': 'openstack',
                    'auth_url': 'http://localhost/identity',
                    'vim_project': {'name': 'test_project'},
                    'auth_cred': {'username': 'test_user',
                                  'password': 'test_password',
                                  'cert_verify': 'True'},
                    'name': 'VIM0',
                    'tenant_id': 'test-project'}}


def get_vim_auth_obj():
    return {'username': 'test_user',
            'password': 'test_password',
            'project_id': None,
            'project_name': 'test_project',
            'cert_verify': 'True',
            'auth_url': 'http://localhost/identity/v3',
            'user_domain_name': 'default',
            'project_domain_name': 'default'}


def get_dummy_vnffgd_obj():
    return {'vnffgd': {'name': 'dummy_vnffgd',
                       'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                       'template': {'vnffgd': vnffgd_tosca_template},
                       'description': 'dummy_vnffgd_description',
                       'template_source': 'onboarded'}}


def get_dummy_vnffgd_obj_inline():
    return {'vnffgd': {'name': 'dummy_vnffgd_inline',
                       'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                       'template': {'vnffgd': vnffgd_tosca_template},
                       'description': 'dummy_vnffgd_description_inline',
                       'template_source': 'inline'}}


def get_dummy_vnffgd_obj_no_description():
    return {'vnffgd': {'name': 'dummy_vnffgd',
                       'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                       'template': {'vnffgd': vnffgd_tosca_template},
                       'template_source': 'onboarded'}}


def get_dummy_vnffgd_obj_no_name():
    return {'vnffgd': {'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                       'template': {'vnffgd': vnffgd_tosca_template},
                       'description': 'dummy_vnffgd_description',
                       'template_source': 'onboarded'}}


def get_dummy_vnffg_obj():
    return {'vnffg': {'description': 'dummy_vnffg_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {'template': vnffgd_tosca_template},
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_vnffg_no_classifier_obj():
    return {'vnffg': {'description': 'dummy_vnffg_no_classifier_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {
                          'template': vnffgd_tosca_no_classifier_template},
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_vnffg_obj_inline():
    return {'vnffg': {'description': 'dummy_vnffg_description_inline',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg_inline',
                      'attributes': {'template': vnffgd_tosca_template},
                      'vnf_mapping': {},
                      'symmetrical': False,
                      'vnffgd_template': vnffgd_tosca_template}}


def get_dummy_vnffg_obj_update_vnffgd_template():
    return {'vnffg': {'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'symmetrical': False,
                      'vnffgd_template': vnffgd_tosca_template_for_update}}


def get_dummy_vnffg_obj_legacy_vnffgd_template():
    return {'vnffg': {'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'symmetrical': False,
                      'vnffgd_template': vnffgd_legacy_template}}


def get_dummy_vnffg_param_obj():
    return {'vnffg': {'description': 'dummy_vnf_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {
                          'template': vnffgd_tosca_param_template,
                          'param_values':
                              yaml.safe_load(vnffg_params)
                      },
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_vnffg_str_param_obj():
    return {'vnffg': {'description': 'dummy_vnf_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {
                          'template': vnffgd_tosca_param_template,
                          'param_values': 'value not dict format'},
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_vnffg_multi_param_obj():
    return {'vnffg': {'description': 'dummy_vnf_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {
                          'template': vnffgd_tosca_multi_param_template,
                          'param_values':
                              yaml.safe_load(vnffg_multi_params)
                      },
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_vnffg_obj_vnf_mapping():
    return {'vnffg': {'description': 'dummy_vnf_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {'template': vnffgd_tosca_template},
                      'vnf_mapping': {
                          'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effe07',
                          'VNF3': '7168062e-9fa1-4203-8cb7-f5c99ff3ee1b'
                      },
                      'symmetrical': False}}


def get_dummy_vnffg_obj_dupl_criteria():
    return {'vnffg': {'description': 'dummy_vnffg_description',
                      'vnffgd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                      'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                      'name': 'dummy_vnffg',
                      'attributes': {'template':
                                     vnffgd_tosca_dupl_criteria_template},
                      'vnf_mapping': {},
                      'symmetrical': False}}


def get_dummy_nsd_obj():
    return {'nsd': {'description': 'dummy_nsd_description',
                    'name': 'dummy_NSD',
                    'tenant_id': '8819a1542a5948b68f94d4be0fd50496',
                    'attributes': {'nsd': nsd_tosca_template},
                    'template_source': 'onboarded'}}


def get_dummy_nsd_obj_inline():
    return {'nsd': {'description': 'dummy_nsd_description_inline',
                    'name': 'dummy_NSD_inline',
                    'tenant_id': '8819a1542a5948b68f94d4be0fd50496',
                    'attributes': {'nsd': nsd_tosca_template},
                    'template_source': 'inline'}}


def get_dummy_ns_obj():
    return {'ns': {'description': 'dummy_ns_description',
                   'id': 'ba6bf017-f6f7-45f1-a280-57b073bf78ea',
                   'nsd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                   'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                   'name': 'dummy_ns',
                   'attributes': {
                       'param_values': {'nsd': {'vl1_name': 'net_mgmt',
                                                'vl2_name': 'net0'}}}}}


def get_dummy_ns_obj_inline():
    return {'ns': {'description': 'dummy_ns_description_inline',
                   'id': 'ff35e3f0-0a11-4071-bce6-279fdf1c8bf9',
                   'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                   'name': 'dummy_ns_inline',
                   'attributes': {
                       'param_values': {'nsd': {'vl1_name': 'net_mgmt',
                                                'vl2_name': 'net0'}}},
                   'nsd_template': nsd_tosca_template}}


def get_dummy_ns_obj_2():
    return {'ns': {'description': 'dummy_ns_description',
                   'id': 'ba6bf017-f6f7-45f1-a280-57b073bf78ea',
                   'nsd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                   'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                   'name': DUMMY_NS_2_NAME,
                   'attributes': {
                       'param_values': {'nsd': {'vl1_name': 'net_mgmt',
                                                'vl2_name': 'net0'}}}}}


def get_dummy_vnf_instance():
    connection_info = get_dummy_vim_connection_info()
    return {'created_at': '', 'deleted': False, 'deleted_at': None,
            'id': 'fake_id', 'instantiated_vnf_info': None,
            'instantiation_state': 'NOT_INSTANTIATED',
            'tenant_id': 'fake_tenant_id', 'updated_at': '',
            'vim_connection_info': [connection_info],
            'vnf_instance_description': 'VNF Description',
            'vnf_instance_name': 'test', 'vnf_product_name': 'Sample VNF',
            'vnf_provider': 'Company', 'vnf_software_version': '1.0',
            'vnfd_id': 'fake_vnfd_id', 'vnfd_version': '1.0'}


def get_dummy_vim_connection_info():
    return {'access_info': {
        'auth_url': 'fake/url',
        'cert_verify': 'False', 'password': 'admin',
        'project_domain_name': 'Default',
        'project_id': None, 'project_name': 'admin',
        'user_domain_name': 'Default', 'username': 'admin'},
        'created_at': '', 'deleted': False, 'deleted_at': '',
        'id': 'fake_id', 'updated_at': '',
        'vim_id': 'fake_vim_id', 'vim_type': 'openstack'}


def get_dummy_grant_hot():
    return str(hot_grant)


def get_dummy_scale_grant_hot():
    return str(hot_scale_grant)


def get_dummy_scale_nest_grant_hot():
    return str(hot_scale_nest_grant)


def get_dummy_scale_initial_hot():
    return str(hot_scale_initial)


def get_dummy_scale_nest_initial_hot():
    return str(hot_scale_nest_initial)



#    Copyright 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import base
from tacker.objects import fields


VNF_SOFTWARE_IMAGE_OPTIONAL_ATTRS = ['metadata']

LOG = logging.getLogger(__name__)


def _metadata_add_to_db(context, id, metadata, max_retries=10):
    for attempt in range(max_retries):
        with db_api.context_manager.writer.using(context):

            new_entries = []
            for key, value in metadata.items():
                new_entries.append({"key": key,
                                    "value": value,
                                    "image_uuid": id})
            if new_entries:
                context.session.execute(
                    models.VnfSoftwareImageMetadata.__table__.insert(None),
                    new_entries)

            return metadata


@db_api.context_manager.writer
def _vnf_sw_image_create(context, values, metadata=None):
    vnf_sw_image = models.VnfSoftwareImage()

    vnf_sw_image.update(values)
    vnf_sw_image.save(context.session)
    vnf_sw_image._metadata = []

    if metadata:
        _metadata_add_to_db(context, vnf_sw_image.id, metadata)
        context.session.expire(vnf_sw_image, ['_metadata'])
        vnf_sw_image._metadata

    return vnf_sw_image


@db_api.context_manager.reader
def _vnf_sw_image_get_by_id(context, id):

    query = api.model_query(context, models.VnfSoftwareImage,
        read_deleted="no").filter_by(id=id).options(joinedload('_metadata'))

    result = query.first()

    if not result:
        raise exceptions.VnfSoftwareImageNotFound(id=id)

    return result


@base.TackerObjectRegistry.register
class VnfSoftwareImage(base.TackerObject, base.TackerPersistentObject):

    ALL_ATTRIBUTES = {
        "softwareImages": {
            'id': ('software_image_id', 'uuid', 'VnfSoftwareImage'),
            'imagePath': ('image_path', 'string', 'VnfSoftwareImage'),
            'diskFormat': ('disk_format', 'string', 'VnfSoftwareImage'),
            'userMetadata/*': ('metadata', 'key_value_pair',
            {"key_column": "key", "value_column": "value",
             "model": "VnfSoftwareImageMetadata"}),
            'size': ('size', 'number', 'VnfSoftwareImage'),
            'createdAt': ('created_at', 'datetime', 'VnfSoftwareImage'),
            'name': ('name', 'string', 'VnfSoftwareImage'),
            'minDisk': ('min_disk', 'number', 'VnfSoftwareImage'),
            'version': ('version', 'string', 'VnfSoftwareImage'),
            'provider': ('provider', 'string', 'VnfSoftwareImage'),
            'minRam': ('min_ram', 'number', 'VnfSoftwareImage'),
            'containerFormat': ('container_format', 'string',
                                'VnfSoftwareImage'),
            "checksum": {
                'hash': ('hash', 'string', 'VnfSoftwareImage'),
                'algorithm': ('algorithm', 'string', 'VnfSoftwareImage')
            }
        }
    }

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())
    SIMPLE_ATTRIBUTES = ['id', 'imagePath', 'diskFormat', 'size',
            'createdAt', 'name', 'minDisk', 'version', 'provider', 'minRam',
            'containerFormat']
    COMPLEX_ATTRIBUTES = ['softwareImages', 'softwareImages/userMetadata',
        'softwareImages/checksum']

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'software_image_id': fields.StringField(nullable=False),
        'flavour_uuid': fields.UUIDField(nullable=False),
        'name': fields.StringField(nullable=True),
        'provider': fields.StringField(nullable=True),
        'version': fields.StringField(nullable=True),
        'algorithm': fields.StringField(nullable=True),
        'hash': fields.StringField(nullable=True),
        'container_format': fields.StringField(nullable=True),
        'disk_format': fields.StringField(nullable=True),
        'min_disk': fields.IntegerField(),
        'min_ram': fields.IntegerField(default=0),
        'size': fields.IntegerField(),
        'image_path': fields.StringField(),
        'metadata': fields.DictOfStringsField(nullable=True)
    }

    @staticmethod
    def _from_db_object(context, vnf_sw_image, db_sw_image,
                        expected_attrs=None):

        vnf_sw_image._context = context
        for key in vnf_sw_image.fields:
            if key in VNF_SOFTWARE_IMAGE_OPTIONAL_ATTRS:
                continue
            else:
                db_key = key

            setattr(vnf_sw_image, key, db_sw_image[db_key])

        vnf_sw_image._extra_attributes_from_db_object(vnf_sw_image,
                          db_sw_image, expected_attrs)

        vnf_sw_image.obj_reset_changes()

        return vnf_sw_image

    @staticmethod
    def _extra_attributes_from_db_object(vnf_sw_image, db_sw_image,
                                         expected_attrs=None):
        """Method to help with migration of extra attributes to objects.

        """
        if expected_attrs is None:
            expected_attrs = []

        if 'metadata' in expected_attrs:
            setattr(vnf_sw_image, 'metadata', db_sw_image['metadetails'])

    def obj_load_attr(self, attrname):
        if not self._context:
            raise exceptions.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())
        if 'id' not in self:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s id %(id)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'id': self.id,
                   })

        self._obj_load_attr(attrname)

    def _obj_load_attr(self, attrname):
        """Internal method for loading attributes from vnf flavour."""

        if attrname in self.fields and attrname != 'id':
            self._load_generic(attrname)
        else:
            # NOTE(nirajsingh): Raise error if non existing field is
            # requested.
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        self.obj_reset_changes([attrname])

    def _load_generic(self, attrname):
        software_image = self.__class__.get_by_id(self._context,
                                              id=self.id,
                                              expected_attrs=attrname)
        if attrname not in software_image:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('loading %s requires recursion') % attrname)

        for field in self.fields:
            if field in software_image and field not in self:
                setattr(self, field, getattr(software_image, field))

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='create',
                                               reason=_('already created'))
        updates = self.obj_get_changes()

        if 'id' not in updates:
            updates['id'] = uuidutils.generate_uuid()
            self.id = updates['id']

        metadata = updates.pop('metadata', None)
        db_sw_image = _vnf_sw_image_create(self._context, updates,
                                           metadata=metadata)
        self._from_db_object(self._context, self, db_sw_image)

    @base.remotable_classmethod
    def get_by_id(cls, context, id, expected_attrs=None):
        db_sw_image = _vnf_sw_image_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_sw_image,
                                   expected_attrs=expected_attrs)

    def _get_user_metadata(self, include_fields=None):
        # Need special handling for field containing key-value pair.
        # If user requests softwareImages/userMetadata/key1 and if
        # softwareImages/userMetadata contains key1=value1, key2=value2,
        # it should return only keys that are requested in include_fields.
        # If user requests only softwareImages/userMetadata, then in that
        # case, it should return all key/value pairs. If any of the requested
        # key is not present, then it will siliently ignore it.
        key = 'softwareImages/userMetadata'
        if key in include_fields or '%s/*' % key in \
                include_fields:
            return self.metadata
        else:
            # Check if user has requested specified keys from
            # softwareImages/userMetadata.
            key_list = []
            special_key = '%s/' % key
            for field in include_fields:
                if field.startswith(special_key):
                    key_list.append(field[len(special_key):])

            data_resp = dict()
            for key_req in key_list:
                if key_req in self.metadata:
                    data_resp[key_req] = self.metadata[key_req]

            if len(key_list) > 0:
                return data_resp

    def to_dict(self, include_fields=None):
        response = dict()
        fields = ['softwareImages/%s' % attribute for attribute in
            self.SIMPLE_ATTRIBUTES]
        to_fields = set(fields).intersection(include_fields)
        for field in to_fields:
            display_field = field.split("/")[-1]
            response[display_field] = getattr(self,
                    self.FLATTEN_ATTRIBUTES[field][0])

        # add checksum
        to_fields = set([key for key in self.FLATTEN_ATTRIBUTES.keys()
            if key.startswith('softwareImages/checksum')])
        checksum = dict()
        to_fields = to_fields.intersection(include_fields)
        for field in to_fields:
            display_field = field.split("/")[-1]
            checksum[display_field] = getattr(self,
                    self.FLATTEN_ATTRIBUTES[field][0])

        if checksum:
            response.update({"checksum": checksum})

        user_metadata = self._get_user_metadata(include_fields)
        if user_metadata is not None:
            response.update({"userMetadata": user_metadata})

        return response


@base.TackerObjectRegistry.register
class VnfSoftwareImagesList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfSoftwareImage')
    }



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import shutil
import tempfile

from oslo_utils import uuidutils

from tacker.tests.functional.sol_v2 import paramgen
from tacker.tests.functional.sol_v2 import utils


zip_file_name = os.path.basename(os.path.abspath(".")) + '.zip'
tmp_dir = tempfile.mkdtemp()
vnfd_id = uuidutils.generate_uuid()

# tacker/tests/etc...
#             /functional/sol_v2/samples/smapleX
image_dir = "../../../../etc/samples/etsi/nfv/common/Files/images/"
image_file = "cirros-0.5.2-x86_64-disk.img"
image_path = os.path.abspath(image_dir + image_file)

utils.make_zip(".", tmp_dir, vnfd_id, image_path)

shutil.move(os.path.join(tmp_dir, zip_file_name), ".")
shutil.rmtree(tmp_dir)

create_req = paramgen.create_vnf_max(vnfd_id)
scaleout_req = paramgen.scaleout_vnf_max()
scalein_req = paramgen.scalein_vnf_max()
terminate_req = paramgen.terminate_vnf_max()

print('#####################################################################\n'
      '# Run pre.py if an error occurs                                     #\n'
      '#  - If an error occurs, run the pre.py script in advance           #\n'
      '#    to create the openstack resource required to run this script.  #\n'
      '# Run post.py when you finish tests                                 #\n'
      '#  - When you no longer need these openstack resources              #\n'
      '#    after testing, run post.py and delete them.                    #\n'
      '# vnfc ids should be changed in heal req files by show vnf manually.#\n'
      '#####################################################################')

net_ids = utils.get_network_ids(['net0', 'net1', 'net_mgmt', 'ft-net0',
    'ft-net1'])
subnet_ids = utils.get_subnet_ids(['subnet0', 'subnet1', 'ft-ipv4-subnet0',
    'ft-ipv6-subnet0', 'ft-ipv4-subnet1', 'ft-ipv6-subnet1'])
port_ids = utils.get_port_ids(['VDU2_CP1-1', 'VDU2_CP1-2'])

instantiate_req = paramgen.instantiate_vnf_max(
    net_ids, subnet_ids, port_ids, "http://localhost/identity/v3")

# fake vnfc id, should be get from show vnf
vnfc_ids = ['VDU1-9300a3cb-bd3b-45e4-9967-095040caf827',
    'VDU2-39681281-e6e6-4179-8898-d9ec70f1642a']
heal_vnfc_req = paramgen.heal_vnf_vnfc_max(vnfc_ids[0])
heal_vnfc_with_omit_all_req = paramgen.heal_vnf_vnfc_max_with_parameter(
    vnfc_ids)
heal_vnfc_with_all_false_req = paramgen.heal_vnf_vnfc_max_with_parameter(
    vnfc_ids, False)
heal_vnfc_with_all_true_req = paramgen.heal_vnf_vnfc_max_with_parameter(
    vnfc_ids, True)
heal_all_with_omit_all_req = paramgen.heal_vnf_all_max_with_parameter()
heal_all_with_all_true_req = paramgen.heal_vnf_all_max_with_parameter(True)
heal_all_with_all_false_req = paramgen.heal_vnf_all_max_with_parameter(False)
change_ext_conn_max_req = paramgen.change_ext_conn_max(net_ids, subnet_ids,
    "http://localhost/identity/v3")
# Only this package have external connectivity.
# So min pattern also use this package.
change_ext_conn_min_req = paramgen.change_ext_conn_min(net_ids, subnet_ids)

with open("create_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(create_req, indent=2))

with open("terminate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(terminate_req, indent=2))

with open("scaleout_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(scaleout_req, indent=2))

with open("scalein_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(scalein_req, indent=2))

with open("instantiate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(instantiate_req, indent=2))

with open("heal_vnfc_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_vnfc_req, indent=2))

with open("heal_vnfc_with_omit_all_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_vnfc_with_omit_all_req, indent=2))

with open("heal_vnfc_with_all_false_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_vnfc_with_all_false_req, indent=2))

with open("heal_vnfc_with_all_true_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_vnfc_with_all_true_req, indent=2))

with open("heal_all_with_omit_all_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_all_with_omit_all_req, indent=2))

with open("heal_all_with_all_true_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_all_with_all_true_req, indent=2))

with open("heal_all_with_all_false_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_all_with_all_false_req, indent=2))

with open("change_ext_conn_max_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(change_ext_conn_max_req, indent=2))

with open("change_ext_conn_min_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(change_ext_conn_min_req, indent=2))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg
from oslo_serialization import jsonutils

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base


CONF = cfg.CONF


class VnfTestToscaScale(base.BaseTackerTest):

    def test_vnf_tosca_scale(self):
        vnf_name = 'test_tosca_vnf_scale_all'
        vnfd_file = 'sample-tosca-scale-all.yaml'
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)
        vnf_id = vnf_instance['vnf']['id']

        # TODO(kanagaraj-manickam) once load-balancer support is enabled,
        # update this logic to validate the scaling
        def _wait(count):
            self.wait_until_vnf_active(
                vnf_id,
                constants.VNF_CIRROS_CREATE_TIMEOUT,
                constants.ACTIVE_SLEEP_TIME)
            vnf = self.client.show_vnf(vnf_id)['vnf']

            # {"VDU1": ["10.0.0.14", "10.0.0.5"]}
            self.assertEqual(count, len(jsonutils.loads(vnf[
                'mgmt_ip_address'])['VDU1']))

        _wait(2)
        # Get nested resources when vnf is in active state
        vnf_details = self.client.list_vnf_resources(vnf_id)['resources']
        resources_list = list()
        for vnf_detail in vnf_details:
            resources_list.append(vnf_detail['name'])
        self.assertIn('VDU1', resources_list)

        self.assertIn('CP1', resources_list)
        self.assertIn('SP1_group', resources_list)

        def _scale(type, count):
            body = {"scale": {'type': type, 'policy': 'SP1'}}
            self.client.scale_vnf(vnf_id, body)
            _wait(count)

        # scale out
        time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)
        _scale('out', 3)

        # scale in
        time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)
        _scale('in', 2)

        # Verifying that as part of SCALE OUT, VNF states  PENDING_SCALE_OUT
        # and ACTIVE occurs and as part of SCALE IN, VNF states
        # PENDING_SCALE_IN and ACTIVE occur.
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_SCALE,
                                    evt_constants.ACTIVE, cnt=2)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_SCALE,
                                    evt_constants.PENDING_SCALE_OUT, cnt=1)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_SCALE,
                                    evt_constants.PENDING_SCALE_IN, cnt=1)
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        # Wait for delete vnf_instance
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from keystoneauth1.identity import v3
from keystoneauth1 import session


class Token(object):
    def __init__(self, username, password, project_name,
                 auth_url, user_domain_name, project_domain_name):
        self.username = username
        self.password = password
        self.auth_url = auth_url
        self.project_name = project_name
        self.user_domain_name = user_domain_name
        self.project_domain_name = project_domain_name

    def create_token(self):
        auth = v3.Password(auth_url=self.auth_url,
                           username=self.username,
                           password=self.password,
                           project_name=self.project_name,
                           user_domain_name=self.user_domain_name,
                           project_domain_name=self.project_domain_name)
        sess = session.Session(auth=auth)
        token_id = sess.auth.get_token(sess)
        return token_id



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
URL = "/zabbix/api_jsonrpc.php"

HEADERS = {'Content-Type': 'application/json-rpc'}
dAUTH_API = {'jsonrpc': "2.0",
             'method': 'user.login',
             'params': {'user': None,
                        'password': None},
             'id': 1,
             'auth': None}

COMP_VALUE = {'greater': '>',
              'less': '<',
              'and greater': '>=',
              'and less': '<=',
              'down': '=0'}


dTEMPLATE_CREATE_API = {'jsonrpc': "2.0", 'method': "template.create",
                        'params': {'host': "", 'groups': {'groupid': 1},
                                   'hosts': []},
                        'id': 1004, 'auth': None}

dITEM_CREATE_API = {'jsonrpc': "2.0",
                    'method': "item.create",
                    'params': {'hostid': None,
                               'interfaceid': 'NULL',
                               'name': "",
                               'key_': "",
                               'type': 0,
                               'value_type': 3,
                               'delay': 1},
                    'id': 1,
                    'auth': None}


dITEM_KEY_VALUE = {'os_agent_info': 'agent.ping',
                   'os_cpu_usage': 'system.cpu.util[,iowait]',
                   'os_cpu_load': 'system.cpu.load[percpu,avg1]',
                   'os_proc_value': 'proc.num[,,run]',
                   'app_status': 'net.tcp.port[ ,*]',
                   'app_memory': 'proc.mem[*,root]'}

dITEM_KEY_COMP = {'os_agent_info': str(
    dITEM_KEY_VALUE['os_agent_info'] + '.nodata(15s)}=1'),
    'os_cpu_usage': str(
        dITEM_KEY_VALUE['os_cpu_usage'] + '.avg(5s)}'),
    'os_cpu_load': str(
        dITEM_KEY_VALUE['os_cpu_load'] + '.avg(5s)}'),
    'os_proc_value': str(
        dITEM_KEY_VALUE['os_proc_value'] + '.avg(5s)}'),
    'app_status': str(
        dITEM_KEY_VALUE['app_status'] + '.last(,5)}'),
    'app_memory': str(
        dITEM_KEY_VALUE['app_memory'] + '.avg(5s)}')}

dITEM_KEY_INFO = {'os_proc_value': {'name': 'process number',
                                    'key_': str(
                                        dITEM_KEY_VALUE['os_proc_value']),
                                    'value_type': 3},
                  'os_cpu_load': {'name': 'cpu load',
                                  'key_': str(
                                      dITEM_KEY_VALUE['os_cpu_load']),
                                  'value_type': 0},
                  'os_cpu_usage': {'name': 'cpu util usage',
                                   'key_': str(
                                       dITEM_KEY_VALUE['os_cpu_usage']),
                                   'value_type': 0},
                  'os_agent_info': {'name': 'Zabbix agent status check',
                                    'key_': str(
                                        dITEM_KEY_VALUE['os_agent_info']),
                                    'value_type': 0},
                  'app_status': {'name': ' service status check',
                                 'key_': str(
                                     dITEM_KEY_VALUE['app_status']),
                                 'value_type': 3},
                  'app_memory': {'name': ' service memory usage',
                                 'key_': str(
                                     dITEM_KEY_VALUE['app_memory']),
                                 'value_type': 3}}


dTRIGGER_CREATE_API = {'jsonrpc': "2.0",
                       'method': "trigger.create",
                       'templateid': None,
                       'auth': None,
                       'id': 1004}

dTRIGGER_INFO = {'itemname': None,
                 'cmdname': None,
                 'cmd-action': None}

dTRIGGER_LIST = {'os_agent_info': [{'description': 'Zabbix agent on '
                                                   '{HOST.NAME} is '
                                                   'unreachable '
                                                   'for 15 seconds',
                                    'expression': '{', 'priority': 3}],
                 'app_status': [{'description': 'Service is down '
                                                'on {HOST.NAME}',
                                 'expression': '{', 'priority': 3}],
                 'app_memory': [{'description': 'Process Memory '
                                                'is lacking '
                                                'on {HOST.NAME}',
                                 'expression': '{', 'priority': 3}],
                 'os_cpu_usage': [{'description': 'Disk I/O is '
                                                  'overloaded '
                                                  'on {HOST.NAME}',
                                   'expression': '{', 'priority': 3}],
                 'os_cpu_load': [{'description': 'Processor load '
                                                 'is too high '
                                                 'on {HOST.NAME}',
                                  'expression': '{', 'priority': 3}],
                 'os_proc_value': [{'description': 'Too many '
                                                   'processes running '
                                                   'on {HOST.NAME}',
                                    'expression': '{', 'priority': 3}]}

dACTION_CREATE_API = {'jsonrpc': "2.0",
                      'method': "action.create",
                      'params': {'name': '',
                                 'eventsource': 0,
                                 'status': 0,
                                 'esc_period': 120,
                                 'def_shortdata': "{TRIGGER.NAME}:"
                                                  "{TRIGGER.STATUS}",
                                 'def_longdata': "{TRIGGER.NAME}: "
                                                 "{TR`IGGER.STATUS}\r\n"
                                                 "Last value: "
                                                 "{ITEM.LASTVALUE]"
                                                 "\r\n\r\n{TRIGGER.URL}",
                                 "filter": {"evaltype": 0,
                                            "conditions": [{'conditiontype': 2,
                                                            'operator': 0,
                                                            'value': None}]},
                                 'operations': [{'operationtype': 1,
                                                 'esc_period': 0,
                                                 'esc_step_from': 1,
                                                 'esc_step_to': 2,
                                                 'evaltype': 0,
                                                 'opcommand': {
                                                     'command': None,
                                                     'type': 0,
                                                     'execute_on': 0},
                                                 'opcommand_hst': [
                                                     {'hostid': None}]
                                                 }]},
                      'auth': None, 'id': 1}

dHOST_CREATE_API = {'jsonrpc': "2.0",
                    'method': "host.create",
                    'params': {'host': 'ubuntu',
                               'interfaces': [
                                   {'type': 1,
                                    'main': 1,
                                    'useip': 1,
                                    'dns': "",
                                    'ip': None,
                                    'port': "10050"}],
                               'templates': [{'templateid': None}],
                               'groups': [{'groupid': None}]},
                    'id': 4, 'auth': None}

dGROUP_GET_API = {'jsonrpc': "2.0", 'method': "hostgroup.get",
                  'params': {'output': 'extend',
                             'filter': {'name': ["Zabbix servers", ]}},
                  'id': 1, 'auth': None}

dGRAPH_CREATE_API = {'jsonrpc': '2.0', 'method': 'graph.create',
                     'params': {'name': None,
                                'width': 900,
                                'height': 200,
                                'gitems': []},
                     'auth': None, 'id': 1004}

dAPP_INFO = {'app_port': None,
             'app_name': None,
             'ssh_username': None,
             'ssh_password': None}

dZBX_INFO = {'zabbix_user': None,
             'zabbix_pass': None,
             'zabbix_ip': None,
             'zabbix_port': None,
             'zabbix_token': None
             }

dACTION_LIST = {'item': None,
                'action': None,
                'trigger_id': None,
                'cmd-action': None}

dVDU_INFO = {'template_id': None,
             'template_name': None,
             'hostid': None,
             'group_id': None,
             'mgmt_ip': None,
             'vdu_id': None,
             'parameters': None,
             'actioninfo': [],
             'appinfo': None,
             'zbx_info': None
             }



# Copyright (c) 2014-2018 China Mobile (SuZhou) Software Technology Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils

from tacker.api.v1.router import APIRouter
from tacker.api.v1.router import Index
from tacker.tests import base
from tacker import wsgi


class TestIndex(base.BaseTestCase):

    def test_index(self):
        request = wsgi.Request.blank(
            "/test/", body=b"{'name': 'tacker'}", method='POST',
            headers={'Content-Type': "application/json"})

        index_cls = Index({"version": "v1"})
        result = index_cls(request)
        expect_body = {'resources': [
            {'collection': 'v1',
             'links': [
                 {'href': 'http://localhost/test/v1',
                  'rel': 'self'}],
             'name': 'version'}]}

        self.assertEqual(expect_body, jsonutils.loads(result.body))
        self.assertEqual('application/json', result.content_type)


@mock.patch('tacker.api.v1.router.APIRouter.factory', return_value=None)
class TestAPIRouter(base.BaseTestCase):

    def test_api_factory(self, factory_mock):
        result = APIRouter().factory({})
        factory_mock.assert_called_once_with({})
        self.assertEqual(None, result)



# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from tacker.tests.functional.sol_v2 import utils

utils.delete_network('ft-net0')
# NOTE: subnet is automatically deleted by network deletion
utils.delete_port('VDU2_CP1-1')
utils.delete_port('VDU2_CP1-2')



# Copyright 2011 VMware, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os
import random

import logging as std_logging

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_service import service
from oslo_utils import excutils
from oslo_utils import importutils

from tacker._i18n import _
from tacker.common import config
from tacker.common import rpc as n_rpc
from tacker import context
from tacker import wsgi


service_opts = [
    cfg.IntOpt('report_interval',
               default=10,
               help=_('Seconds between running components report states')),
    cfg.IntOpt('periodic_interval',
               default=40,
               help=_('Seconds between running periodic tasks')),
    cfg.IntOpt('api_workers',
               default=0,
               help=_('Number of separate worker processes for service')),
    cfg.IntOpt('periodic_fuzzy_delay',
               default=5,
               help=_('Range of seconds to randomly delay when starting the '
                      'periodic task scheduler to reduce stampeding. '
                      '(Disable by setting to 0)')),
]
CONF = cfg.CONF
CONF.register_opts(service_opts)


def config_opts():
    return [(None, service_opts)]


LOG = logging.getLogger(__name__)


class WsgiService(service.ServiceBase):
    """Base class for WSGI based services.

    For each api you define, you must also define these flags:
    :<api>_listen: The address on which to listen
    :<api>_listen_port: The port on which to listen

    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.wsgi_app = None

    def start(self):
        self.wsgi_app = _run_wsgi(self.app_name)

    def wait(self):
        if self.wsgi_app:
            self.wsgi_app.wait()

    def stop(self):
        pass

    def reset(self):
        pass


class TackerApiService(WsgiService):
    """Class for tacker-api service."""

    @classmethod
    def create(cls, app_name='tacker'):

        # Setup logging early
        config.setup_logging(cfg.CONF)
        # Dump the initial option values
        cfg.CONF.log_opt_values(LOG, std_logging.DEBUG)
        service = cls(app_name)
        return service


def serve_wsgi(cls):

    try:
        service = cls.create()
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.exception('Unrecoverable error: please check log '
                          'for details.')

    return service


def _run_wsgi(app_name):
    app = config.load_paste_app(app_name)
    if not app:
        LOG.error('No known API applications configured.')
        return
    server = wsgi.Server("Tacker")
    server.start(app, cfg.CONF.bind_port, cfg.CONF.bind_host)
    # Dump all option values here after all options are parsed
    cfg.CONF.log_opt_values(LOG, std_logging.DEBUG)
    LOG.info("Tacker service started, listening on %(host)s:%(port)s",
             {'host': cfg.CONF.bind_host,
              'port': cfg.CONF.bind_port})
    return server


class Service(n_rpc.Service):
    """Service object for binaries running on hosts.

    A service takes a manager and enables rpc by listening to queues based
    on topic. It also periodically runs tasks on the manager.
    """

    def __init__(self, host, binary, topic, manager, report_interval=None,
                 periodic_interval=None, periodic_fuzzy_delay=None,
                 *args, **kwargs):

        self.binary = binary
        self.manager_class_name = manager
        manager_class = importutils.import_class(self.manager_class_name)
        self.manager = manager_class(host=host, *args, **kwargs)
        self.report_interval = report_interval
        self.periodic_interval = periodic_interval
        self.periodic_fuzzy_delay = periodic_fuzzy_delay
        self.saved_args, self.saved_kwargs = args, kwargs
        self.timers = []
        super(Service, self).__init__(host, topic, manager=self.manager)

    def start(self):
        self.manager.start()
        self.manager.init_host()
        super(Service, self).start()
        if self.report_interval:
            pulse = loopingcall.FixedIntervalLoopingCall(self.report_state)
            pulse.start(interval=self.report_interval,
                        initial_delay=self.report_interval)
            self.timers.append(pulse)

        if self.periodic_interval:
            if self.periodic_fuzzy_delay:
                initial_delay = random.randint(0, self.periodic_fuzzy_delay)
            else:
                initial_delay = None

            periodic = loopingcall.FixedIntervalLoopingCall(
                self.periodic_tasks)
            periodic.start(interval=self.periodic_interval,
                           initial_delay=initial_delay)
            self.timers.append(periodic)
        self.manager.after_start()

    def __getattr__(self, key):
        manager = self.__dict__.get('manager', None)
        return getattr(manager, key)

    @classmethod
    def create(cls, host=None, binary=None, topic=None, manager=None,
               report_interval=None, periodic_interval=None,
               periodic_fuzzy_delay=None):
        """Instantiates class and passes back application object.

        :param host: defaults to cfg.CONF.host
        :param binary: defaults to basename of executable
        :param topic: defaults to bin_name - 'tacker-' part
        :param manager: defaults to cfg.CONF.<topic>_manager
        :param report_interval: defaults to cfg.CONF.report_interval
        :param periodic_interval: defaults to cfg.CONF.periodic_interval
        :param periodic_fuzzy_delay: defaults to cfg.CONF.periodic_fuzzy_delay

        """
        if not host:
            host = cfg.CONF.host
        if not binary:
            binary = os.path.basename(inspect.stack()[-1][1])
        if not topic:
            topic = binary.rpartition('tacker-')[2]
            topic = topic.replace("-", "_")
        if not manager:
            manager = cfg.CONF.get('%s_manager' % topic, None)
        if report_interval is None:
            report_interval = cfg.CONF.report_interval
        if periodic_interval is None:
            periodic_interval = cfg.CONF.periodic_interval
        if periodic_fuzzy_delay is None:
            periodic_fuzzy_delay = cfg.CONF.periodic_fuzzy_delay
        service_obj = cls(host, binary, topic, manager,
                          report_interval=report_interval,
                          periodic_interval=periodic_interval,
                          periodic_fuzzy_delay=periodic_fuzzy_delay)

        return service_obj

    def kill(self):
        """Destroy the service object."""
        self.stop()

    def stop(self):
        self.manager.stop()
        super(Service, self).stop()
        for x in self.timers:
            try:
                x.stop()
            except Exception:
                LOG.exception("Exception occurs when timer stops")
        self.timers = []

    def wait(self):
        super(Service, self).wait()
        for x in self.timers:
            try:
                x.wait()
            except Exception:
                LOG.exception("Exception occurs when waiting for timer")

    def reset(self):
        config.reset_service()

    def periodic_tasks(self, raise_on_error=False):
        """Tasks to be run at a periodic interval."""
        ctxt = context.get_admin_context()
        self.manager.periodic_tasks(ctxt, raise_on_error=raise_on_error)

    def report_state(self):
        """Update the state of this service."""
        # Todo(gongysh) report state to neutron server
        pass



#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json
import testtools
import time
import yaml

from blazarclient import exception
from oslo_serialization import jsonutils
from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file


HYPERVISORS = []


def hypervisors():
    global HYPERVISORS
    client = base.BaseTackerTest.novaclient()
    hypervisor_list = client.hypervisors.list()
    for hypervisor in hypervisor_list:
        if hypervisor.running_vms == 0:
            HYPERVISORS.append(hypervisor)
    return HYPERVISORS


class VnfTestReservationMonitor(base.BaseTackerTest):

    def _test_vnf_tosca_reservation(self, vnfd_file, vnf_name,
                              lease_id, param_values=None):
        input_yaml = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)

        # TODO(niraj-singh): It's not possible to pass parameters through
        # parameter file due to Bug #1799683. Once this bug is fixed, no need
        # to update vnfd yaml.
        vdu_prop = tosca_dict['topology_template']['node_templates']['VDU1']
        vdu_prop['properties']['flavor'] = param_values.get('flavor')
        vdu_prop['properties']['reservation_metadata']['id'] =\
            param_values.get('server_group_id')
        vdu_prop['properties']['reservation_metadata']['resource_type'] =\
            param_values.get('resource_type')
        policies = tosca_dict['topology_template']['policies']
        policies[0]['RSV']['reservation']['properties']['lease_id'] =\
            param_values.get('lease_id')

        tosca_arg = {'vnfd': {'name': vnf_name,
                              'attributes': {'vnfd': tosca_dict}}}
        blazarclient = self.blazarclient()

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Create vnf with vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']
        self.addCleanup(self.client.delete_vnfd, vnfd_id)

        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)

        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        vnf_id = vnf_instance['vnf']['id']

        def _get_reservation_policy(vnfd_dict):
            policies = vnfd_dict['topology_template'].get('policies', [])
            res_policy = dict()
            for policy_dict in policies:
                for name, policy in policy_dict.items():
                    if policy['type'] == constants.POLICY_RESERVATION:
                        reservations = policy['reservation']
                        for reserv_key, reserv_value in reservations.items():
                            if reserv_key == 'properties':
                                continue
                            for policy_action in reserv_value:
                                res_policy[reserv_key] = policy_action
            return res_policy

        def _check_lease_event_status():
            lease_event_status = _verify_and_get_lease_event_status()
            self.assertEqual(lease_event_status, constants.LEASE_EVENT_STATUS,
                             "Lease %(lease_id)s with status %(status)s is"
                             " expected to be %(target)s" %
                             {"lease_id": lease_id,
                              "status": lease_event_status,
                              "target": constants.LEASE_EVENT_STATUS})

        def _verify_and_get_lease_event_status():
            start_time = int(time.time())
            while ((int(time.time()) - start_time) <
                   constants.LEASE_CHECK_EVENT_TIMEOUT):
                lease_detail = blazarclient.lease.get(lease_id)
                lease_events = lease_detail.get('events')
                for event in lease_events:
                    lease_event_status = event.get('status')
                    if ((event.get('event_type') ==
                         constants.START_LEASE_EVET_TYPE) and (
                            lease_event_status ==
                            constants.LEASE_EVENT_STATUS)):
                        return lease_event_status
                time.sleep(constants.LEASE_CHECK_SLEEP_TIME)

        def _wait_vnf_active_and_assert_vdu_count(vdu_count, scale_type=None):
            self.wait_until_vnf_active(
                vnf_id,
                constants.VNF_CIRROS_CREATE_TIMEOUT,
                constants.ACTIVE_SLEEP_TIME)

            vnf = self.client.show_vnf(vnf_id)['vnf']
            # {"VDU1": ["10.0.0.14", "10.0.0.5"]}
            if scale_type == 'scaling-in' and vdu_count == 0:
                self.assertFalse(jsonutils.loads(vnf.get('mgmt_ip_address',
                                                         '{}')),
                                 "Once scaling-in completes, mgmt_ip_address"
                                 " should be empty.")
            elif scale_type == 'scaling-out':
                self.assertEqual(vdu_count, len(json.loads(
                    vnf['mgmt_ip_address'])['VDU1']))
            elif vdu_count == 0 and scale_type is None:
                self.assertIsNone(vnf['mgmt_ip_address'])

        reservation_policy = _get_reservation_policy(tosca_dict)
        _wait_vnf_active_and_assert_vdu_count(0)

        # trigger alarm for start action
        start_action = reservation_policy.get('start_actions')
        scaling_out_action = start_action + '-out'
        _check_lease_event_status()
        # scaling-out action
        self.trigger_vnf(vnf_id, 'start_actions', scaling_out_action)
        time.sleep(constants.SCALE_SLEEP_TIME)
        # checking VDU's count after scaling out
        _wait_vnf_active_and_assert_vdu_count(2, scale_type='scaling-out')
        time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)

        # trigger alarm for before end action
        before_end_action = reservation_policy.get('before_end_actions')
        scaling_in_action = before_end_action + '-in'

        # scaling-in action
        self.trigger_vnf(vnf_id, 'before_end_actions', scaling_in_action)
        time.sleep(constants.SCALE_SLEEP_TIME)
        # checking VDU's count after scaling in
        _wait_vnf_active_and_assert_vdu_count(0, scale_type='scaling-in')

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_SCALE,
            evt_constants.ACTIVE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_SCALE,
            evt_constants.PENDING_SCALE_OUT, cnt=1)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_SCALE,
            evt_constants.PENDING_SCALE_IN, cnt=1)

        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, (
                "Failed to delete vnf %s after the reservation test" % vnf_id)
        # Wait for delete vnf_instance
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)

    def _get_instance_reservation(self):
        blazarclient = self.blazarclient()
        reservations = [{'disk_gb': 0,
                         'vcpus': 1, 'memory_mb': 1,
                         'amount': 2, 'affinity': True,
                         'resource_properties': '',
                         'resource_type': 'virtual:instance'}]
        events = []

        start_date = (datetime.datetime.utcnow() + datetime.timedelta(
            minutes=2)).strftime("%Y-%m-%d %H:%M")
        end_date = (datetime.datetime.utcnow() + datetime.timedelta(
            minutes=30)).strftime("%Y-%m-%d %H:%M")
        host_added = False
        for hypervisor in HYPERVISORS:
            try:
                blazar_host = blazarclient.host.create(
                    hypervisor.hypervisor_hostname)
                host_added = True
                break
            except exception.BlazarClientException:
                pass

        if not host_added:
            self.skipTest("Skip test as Blazar failed to create host from one"
                          " of available hypervisors '%s' as it found some"
                          " instances were running on them" %
                          ",".join([hypervisor.hypervisor_hostname
                                    for hypervisor in HYPERVISORS]))
        instance_reservation = blazarclient.lease.create(
            'test-reservation', start_date, end_date, reservations, events)

        self.addCleanup(
            blazarclient.host.delete, blazar_host.get('id'))
        self.addCleanup(
            blazarclient.lease.delete, instance_reservation['id'])

        return instance_reservation, blazar_host

    @testtools.skipIf(len(hypervisors()) == 0,
                      'Skip test as there are no'
                      ' hypervisors available in nova')
    def test_vnf_alarm_scale_with_instance_reservation(self):
        instance_reservation, blazar_host = self._get_instance_reservation()
        lease_id = str(instance_reservation['reservations'][0]['lease_id'])
        flavor_id = str(instance_reservation['reservations'][0]['flavor_id'])
        server_group_id = str(
            instance_reservation['reservations'][0]['server_group_id'])
        param_to_create_lease = {'lease_id': lease_id,
                                 'flavor': flavor_id,
                                 'server_group_id': server_group_id,
                                 'resource_type': 'virtual_instance'}
        self._test_vnf_tosca_reservation(
            'sample-tosca-vnfd-instance-reservation.yaml',
            'VNFD1', lease_id, param_to_create_lease)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import threading

from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils

from tacker.sol_refactored.api import api_version
from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import http_client
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored import objects


LOG = logging.getLogger(__name__)

CONF = config.CONF

TEST_NOTIFICATION_TIMEOUT = 20  # seconds


def get_subsc(context, subsc_id):
    subsc = objects.LccnSubscriptionV2.get_by_id(context, subsc_id)
    if subsc is None:
        raise sol_ex.LccnSubscriptionNotFound(subsc_id=subsc_id)
    return subsc


def get_subsc_all(context):
    return objects.LccnSubscriptionV2.get_all(context)


def subsc_href(subsc_id, endpoint):
    return "{}/vnflcm/v2/subscriptions/{}".format(endpoint, subsc_id)


def _get_notification_auth_handle(subsc):
    if not subsc.obj_attr_is_set('authentication'):
        return http_client.NoAuthHandle()
    elif subsc.authentication.obj_attr_is_set('paramsBasic'):
        param = subsc.authentication.paramsBasic
        return http_client.BasicAuthHandle(param.userName, param.password)
    elif subsc.authentication.obj_attr_is_set(
            'paramsOauth2ClientCredentials'):
        param = subsc.authentication.paramsOauth2ClientCredentials
        return http_client.OAuth2AuthHandle(None,
            param.tokenEndpoint, param.clientId, param.clientPassword)

    # not reach here


def async_call(func):
    def inner(*args, **kwargs):
        th = threading.Thread(target=func, args=args,
                kwargs=kwargs, daemon=True)
        th.start()
    return inner


@async_call
def send_notification(subsc, notif_data):
    auth_handle = _get_notification_auth_handle(subsc)
    client = http_client.HttpClient(auth_handle,
        version=api_version.CURRENT_VERSION)

    url = subsc.callbackUri
    try:
        resp, body = client.do_request(url, "POST", body=notif_data)
        if resp.status_code != 204:
            LOG.error("send_notification failed: %d" % resp.status_code)
    except Exception:
        # it may occur if test_notification was not executed.
        LOG.exception("send_notification failed")


def test_notification(subsc):
    auth_handle = _get_notification_auth_handle(subsc)
    client = http_client.HttpClient(auth_handle,
        version=api_version.CURRENT_VERSION,
        timeout=TEST_NOTIFICATION_TIMEOUT)

    url = subsc.callbackUri
    try:
        resp, _ = client.do_request(url, "GET")
        if resp.status_code != 204:
            raise sol_ex.TestNotificationFailed()
    except Exception:
        # any sort of error is considered. avoid 500 error.
        raise sol_ex.TestNotificationFailed()


def match_version(version, inst):
    # - vnfSoftwareVersion 1
    # - vnfdVersions 0..N
    if version.vnfSoftwareVersion != inst.vnfSoftwareVersion:
        return False

    if not version.obj_attr_is_set('vnfdVersions'):
        # OK, no more check necessary.
        return True

    return inst.vnfdVersion in version.vnfdVersions


def match_products_per_provider(products, inst):
    # - vnfProvider 1
    # - vnfProducts 0..N
    #   - vnfProductName 1
    #   - versions 0..N
    #     - vnfSoftwareVersion 1
    #     - vnfdVersions 0..N
    if products.vnfProvider != inst.vnfProvider:
        return False

    if not products.obj_attr_is_set('vnfProducts'):
        # OK, no more check necessary.
        return True

    for product in products.vnfProducts:
        if product.vnfProductName == inst.vnfProductName:
            if not product.obj_attr_is_set('versions'):
                # OK, no more check necessary.
                return True
            for ver in product.versions:
                if match_version(ver, inst):
                    # OK, match.
                    return True
    # no match
    return False


def match_inst_subsc_filter(inst_filter, inst):
    # inst_filter: VnfInstanceSubscriptionFilter
    # - vnfdIds 0..N
    # - VnfProductsFromProviders 0..N
    # - vnfInstanceIds 0..N
    # - vnfInstanceNames 0..N
    if inst_filter.obj_attr_is_set('vnfdIds'):
        if inst.vnfdId not in inst_filter.vnfdIds:
            return False

    if inst_filter.obj_attr_is_set('vnfProductsFromProviders'):
        products_providers = inst_filter.vnfProductsFromProviders
        match = False
        for products in products_providers:
            if match_products_per_provider(products, inst):
                match = True
                break
        if not match:
            # no match found
            return False

    if inst_filter.obj_attr_is_set('vnfInstanceIds'):
        if inst.id not in inst_filter.vnfInstanceIds:
            return False

    if inst_filter.obj_attr_is_set('vnfInstanceNames'):
        if inst.vnfInstanceNames not in inst_filter.vnfInstanceNames:
            return False

    return True


def get_inst_create_subscs(context, inst):
    return get_matched_subscs(context, inst,
            'VnfIdentifierCreationNotification', None, None)


def get_inst_delete_subscs(context, inst):
    return get_matched_subscs(context, inst,
            'VnfIdentifierDeletionNotification', None, None)


def get_lcmocc_subscs(context, lcmocc, inst):
    return get_matched_subscs(context, inst,
            'VnfLcmOperationOccurrenceNotification',
            lcmocc.operation, lcmocc.operationState)


def get_matched_subscs(context, inst, notif_type, op_type, op_status):
    subscs = []
    for subsc in get_subsc_all(context):
        # subsc: LccnSubscription

        if not subsc.obj_attr_is_set('filter'):
            # no filter. get it.
            subscs.append(subsc)
            continue

        # subsc.fulter: LifecycleChangeNotificationsFilter
        # - vnfInstanceSubscriptionFilter 0..1
        # - notificationTypes 0..N
        # - operationTypes 0..N
        # - operationStatus 0..N
        if subsc.filter.obj_attr_is_set('vnfInstanceSubscriptionFilter'):
            inst_filter = subsc.filter.vnfInstanceSubscriptionFilter
            if not match_inst_subsc_filter(inst_filter, inst):
                continue

        if subsc.filter.obj_attr_is_set('notificationTypes'):
            if notif_type not in subsc.filter.notificationTypes:
                continue

        if (op_type is not None and
                subsc.filter.obj_attr_is_set('operationTypes')):
            if op_type not in subsc.filter.operationTypes:
                continue

        if (op_status is not None and
                subsc.filter.obj_attr_is_set('operationStatus')):
            if op_status not in subsc.filter.operationStatus:
                continue

        # OK, matched
        subscs.append(subsc)

    return subscs


def make_create_inst_notif_data(subsc, inst, endpoint):
    notif_data = objects.VnfIdentifierCreationNotificationV2(
        id=uuidutils.generate_uuid(),
        notificationType="VnfIdentifierCreationNotification",
        subscriptionId=subsc.id,
        timeStamp=timeutils.utcnow(),
        vnfInstanceId=inst.id,
        _links=objects.LccnLinksV2(
            vnfInstance=objects.NotificationLink(
                href=inst_utils.inst_href(inst.id, endpoint)),
            subscription=objects.NotificationLink(
                href=subsc_href(subsc.id, endpoint))
        )
        # vnfLcmOpOcc: is not necessary
    )
    return notif_data


def make_delete_inst_notif_data(subsc, inst, endpoint):
    notif_data = objects.VnfIdentifierDeletionNotificationV2(
        id=uuidutils.generate_uuid(),
        notificationType="VnfIdentifierDeletionNotification",
        subscriptionId=subsc.id,
        timeStamp=timeutils.utcnow(),
        vnfInstanceId=inst.id,
        _links=objects.LccnLinksV2(
            vnfInstance=objects.NotificationLink(
                href=inst_utils.inst_href(inst.id, endpoint)),
            subscription=objects.NotificationLink(
                href=subsc_href(subsc.id, endpoint))
        )
        # vnfLcmOpOcc: is not necessary
    )
    return notif_data



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from unittest import mock

from oslo_utils import uuidutils

from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.db.nfvo import nfvo_db
from tacker.db.vnfm import vnfm_db
from tacker.plugins.common import constants
from tacker.tests.unit.db import base as db_base
from tacker.tests.unit.db import utils
from tacker.vnfm import plugin


class FakeCVNFMonitor(mock.Mock):
    pass


class FakePlugin(mock.Mock):
    pass


class FakeK8SVimClient(mock.Mock):
    pass


class TestCVNFMPlugin(db_base.SqlTestCase):
    def setUp(self):
        super(TestCVNFMPlugin, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self._mock_vim_client()
        self._stub_get_vim()
        self._mock_vnf_monitor()
        self._mock_vnf_maintenance_monitor()
        self._mock_vnf_maintenance_plugin()
        self._insert_dummy_vim()
        self.vnfm_plugin = plugin.VNFMPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        mock.patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._mgmt_driver_name',
                   return_value='noop').start()
        self.create = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                 'kubernetes_driver.Kubernetes.create',
                                 return_value=uuidutils.
                                 generate_uuid()).start()
        self.create_wait = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                      'kubernetes_driver.Kubernetes.'
                                      'create_wait').start()
        self.update = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                 'kubernetes_driver.Kubernetes.update').start()
        self.update_wait = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                      'kubernetes_driver.Kubernetes.'
                                      'update_wait').start()
        self.delete = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                 'kubernetes_driver.Kubernetes.delete').start()
        self.delete_wait = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                      'kubernetes_driver.Kubernetes.'
                                      'delete_wait').start()
        self.scale = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                'kubernetes_driver.Kubernetes.scale',
                                return_value=uuidutils.generate_uuid()).start()
        self.scale_wait = mock.patch('tacker.vnfm.infra_drivers.kubernetes.'
                                     'kubernetes_driver.Kubernetes.scale_wait',
                                     return_value=uuidutils.
                                     generate_uuid()).start()

        def _fake_spawn(func, *args, **kwargs):
            func(*args, **kwargs)

        mock.patch.object(self.vnfm_plugin, 'spawn_n',
                          _fake_spawn).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()

    def _mock_vim_client(self):
        self.vim_client = mock.Mock(wraps=FakeK8SVimClient())
        fake_vim_client = mock.Mock()
        fake_vim_client.return_value = self.vim_client
        self._mock(
            'tacker.vnfm.vim_client.VimClient', fake_vim_client)

    def _stub_get_vim(self):
        vim_obj = {'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'vim_name': 'fake_vim',
                   'vim_auth': {'auth_url': 'http://localhost:6443',
                                'password': 'test_pw', 'username': 'test_user',
                                'project_name': 'test_project',
                                'ssl_ca_cert': None},
                   'vim_type': 'kubernetes'}
        self.vim_client.get_vim.return_value = vim_obj

    def _mock_vnf_monitor(self):
        self._vnf_monitor = mock.Mock(wraps=FakeCVNFMonitor())
        fake_vnf_monitor = mock.Mock()
        fake_vnf_monitor.return_value = self._vnf_monitor
        self._mock(
            'tacker.vnfm.monitor.VNFMonitor', fake_vnf_monitor)

    def _mock_vnf_maintenance_monitor(self):
        self._vnf_maintenance_mon = mock.Mock(wraps=FakeCVNFMonitor())
        fake_vnf_maintenance_monitor = mock.Mock()
        fake_vnf_maintenance_monitor.return_value = self._vnf_maintenance_mon
        self._mock(
            'tacker.vnfm.monitor.VNFMaintenanceAlarmMonitor',
            fake_vnf_maintenance_monitor)

    def _mock_vnf_maintenance_plugin(self):
        self._vnf_maintenance_plugin = mock.Mock(wraps=FakePlugin())
        fake_vnf_maintenance_plugin = mock.Mock()
        fake_vnf_maintenance_plugin.return_value = self._vnf_maintenance_plugin
        self._mock(
            'tacker.plugins.fenix.FenixPlugin',
            fake_vnf_maintenance_plugin)

    def _insert_dummy_vnf_template(self):
        session = self.context.session
        vnf_template = vnfm_db.VNFD(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template_source='onboarded',
            deleted_at=datetime.min)
        session.add(vnf_template)
        session.flush()
        return vnf_template

    def _insert_dummy_vnf_template_inline(self):
        session = self.context.session
        vnf_template = vnfm_db.VNFD(
            id='d58bcc4e-d0cf-11e6-bf26-cec0c932ce01',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='tmpl-koeak4tqgoqo8cr4-dummy_inline_vnf',
            description='inline_fake_template_description',
            deleted_at=datetime.min,
            template_source='inline')
        session.add(vnf_template)
        session.flush()
        return vnf_template

    def _insert_dummy_vim(self):
        pass
        session = self.context.session
        vim_db = nfvo_db.Vim(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vim',
            description='fake_vim_description',
            type='kubernetes',
            status='Active',
            deleted_at=datetime.min,
            placement_attr={'regions': ['default', 'kube-public',
                                        'kube-system']})
        vim_auth_db = nfvo_db.VimAuth(
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            password='encrypted_pw',
            auth_url='http://localhost:6443',
            vim_project={'name': 'test_project'},
            auth_cred={'auth_url': 'https://localhost:6443',
                       'username': 'admin',
                       'bearer_token': None,
                       'ssl_ca_cert': 'test',
                       'project_name': 'default',
                       'type': 'kubernetes'})
        session.add(vim_db)
        session.add(vim_auth_db)
        session.flush()

    def test_create_cvnf_with_vnfd(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('created_at', result)
        self.assertIn('updated_at', result)
        self.assertEqual('ACTIVE', result['status'])
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_CREATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    @mock.patch('tacker.vnfm.plugin.VNFMPlugin.create_vnfd')
    def test_create_cvnf_from_template(self, mock_create_vnfd):
        self._insert_dummy_vnf_template_inline()
        mock_create_vnfd.return_value = {'id':
                                         'd58bcc4e-d0cf-11e6-bf26'
                                         '-cec0c932ce01'}
        vnf_obj = utils.get_dummy_inline_cvnf_obj()
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('created_at', result)
        self.assertIn('updated_at', result)
        self.assertEqual('ACTIVE', result['status'])
        mock_create_vnfd.assert_called_once_with(mock.ANY, mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_CREATE,
            res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    def test_delete_vnf(self):
        pass

    def test_update_vnf(self):
        pass

    def _test_scale_vnf(self, type):
        pass

    def test_scale_vnf_out(self):
        pass

    def test_scale_vnf_in(self):
        pass
